from Orbaplaw.OrbitalAlignment import OrbitalAlignment
OrbitalAlignment = OrbitalAlignment.OrbitalAlignment

from Orbaplaw.OrbitalAlignment import FragmentAlignment
FragmentAlignment = FragmentAlignment.FragmentAlignment

from Orbaplaw.OrbitalAlignment import SpinAlignment
SpinAlignment = SpinAlignment.SpinAlignment



import numpy as np
import Maniverse as mv

def Fock_func(U, order, Eref):
	Fref = np.diag(Eref)
	F2ref = np.diag(Eref**2)
	F = U.T @ Fref @ U
	DiagF = np.diag(np.diag(F))
	F2 = U.T @ F2ref @ U
	L = np.trace(F2) - np.linalg.norm(DiagF) ** 2
	Ge = 2 * F2ref @ U - 4 * Fref @ U @ DiagF
	twoF2ref = 2 * F2ref
	fourFref = 4 * Fref
	eightFrefU = 8 * Fref @ U
	UTFref = U.T @ Fref
	def He(v):
		return v
	if order == 2:
		def He(v):
			Hv = twoF2ref @ v - fourFref @ v @ DiagF - eightFrefU @ np.diag(np.diag(UTFref @ v))
			return Hv
	return L, Ge, He

def Fock(Eref):
	M = mv.Orthogonal(np.eye(len(Eref)), True)
	def func(U, order):
		L, Ge, He = Fock_func(U, order, Eref)
		return L, Ge, He
	L = 0
	tr_setting = mv.TrustRegionSetting()
	tol0 = 1e-8 * M.getDimension()
	tol1 = 1e-6 * M.getDimension()
	tol2 = 10
	mv.TrustRegion(
			func, tr_setting, (tol0, tol1, tol2),
			0.001, 1, 1000, L, M, 1
	)
	return M.P



import numpy as np
import scipy.linalg as sl
import itertools as it
import copy as cp
from Orbaplaw import Integrals as eint
from Orbaplaw import Localization as loc


class pNBO:
	Combination=None
	pNHOs=None
	Occupancy=None
	Vector=None
	pNMEB=None

class pNHO:
	Fragment=None
	pNBO=None
	Multi=None
	Vector=None
	Siblings=None

class NBO:
	Combination=None
	NHOs=None
	Occupancy=None
	Vector=None
	Multi=None

class NHO:
	Fragment=None
	NBO=None
	Occupancy=None
	Vector=None

def getMatrix(pnbos): # Or pnhos.
	matrix=np.zeros((len(pnbos[0].Vector),len(pnbos)))
	for ipnbo,pnbo in zip(range(len(pnbos)),pnbos):
		matrix[:,ipnbo]=pnbo.Vector
	return matrix

def setMatrix(pnbos,matrix): # Or pnhos.
	for ipnbo,pnbo in zip(range(len(pnbos)),pnbos):
		pnbo.Vector=matrix[:,ipnbo]

def SortpNBO(pnbos,feature): # Or pnhos.
	pnbos_features=dict() # Sorting pNBOs by features, [feature] -> [[pNBO]]. "feature" can be "Combination", "Fragment", etc.
	pnbo_feature=None
	for pnbo in pnbos:
		match feature:
			case "Combination":
				pnbo_feature=pnbo.Combination
			case "Fragment":
				pnbo_feature=pnbo.Fragment
		if pnbo_feature in pnbos_features.keys():
			pnbos_features[pnbo_feature].append(pnbo)
		else:
			pnbos_features[pnbo_feature]=[pnbo]
	return pnbos_features


def generateNaturalBondOrbital(basis_indices_by_frag,P,C,S,maxnfrags,maxnnbos,occ_thres,multi_thres,pdeg_thres,deg_thres):
	# P - The NAO-based density matrix.
	# C - The NAO coefficient matrix in AO basis set.
	# S - The AO-based overlap matrix.

	nbasis=P.shape[0]
	Pr=cp.deepcopy(P) # Residual of matrix P.
	pnbos=[]
	pnhos=[]

	# Searching for pNBOs and pNHOs.
	for nfrags in range(maxnfrags+1):
		combs=list(it.combinations(range(len(basis_indices_by_frag)),nfrags)) # Combinations of fragments.
		for comb in combs:
			bic=[] # Basis indices of this combination, [fragment] -> [[basis]].
			for frag in comb:
				bic.append(basis_indices_by_frag[frag])
			bicf=sum(bic,[]) # Flattened basis indices of this combination, [fragment -> basis]
			Pblock=Pr[np.ix_(bicf,bicf)] # Block of P matrix belonging to this combination
			Nblock,Hblock=np.linalg.eigh(Pblock) # The related occupation and orbitals.
			Nblock=Nblock[::-1] # Putting the eigenvalues and eigenvectors in decreasing order.
			Hblock=np.fliplr(Hblock)
			for ieigen in range(len(bicf)):
				if Nblock[ieigen]<occ_thres: # Ignoring all small occupancies.
					Nblock[ieigen]=0
				else:
					pnbo=pNBO()
					pnbo.Combination=comb
					pnbo.pNHOs=[]
					pnbo.Occupancy=Nblock[ieigen]
					pnbo.Vector=np.zeros(nbasis)
					pnbo.Vector[bicf]=Hblock[:,ieigen]
					for frag,bif in zip(comb,bic):
						pnho=pNHO()
						pnho.Fragment=frag
						pnho.pNBO=pnbo
						pnho.Multi=0
						pnho.Vector=np.zeros(nbasis)
						pnho.Siblings=[]
						pnho.Explored=False
						pnbo.pNHOs.append(pnho)
						pnhos.append(pnho)
					pnbos.append(pnbo)
			Pr[np.ix_(bicf,bicf)]-=Hblock@np.diag(Nblock)@Hblock.T
			if len(pnbos)>=maxnnbos:
				break
		
	# Finding degenerate pNBOs and partially localizing them.
	pnbos_combs=SortpNBO(pnbos,"Combination") # Sorting pNBOs by combinations, [combination : pNBO].
	eigenlists=[] # List of indices of degenerate pNBOs, [degenerate group] -> [[pNBO]].
	for comb,pnbos_comb in pnbos_combs.items():
		if len(comb)==1:
			continue
		eigenvalue=114514
		eigenlist=[] # List of indices of degenerate pNBOs in this degenerate group, [pNBO].
		for pnbo in pnbos_comb:
			if not np.isclose(eigenvalue,pnbo.Occupancy,atol=pdeg_thres) and eigenlist!=[]:
				eigenlists.append(eigenlist) # Recording this degenerate group.
				eigenlist=[] # Clean the degenerate pNBO list.
			eigenvalue=pnbo.Occupancy # Recording the eigenvalue and the index of the current pNBO.
			eigenlist.append(pnbo)
			if pnbo is pnbos_comb[-1]: # One group of degenerate pNBOs is found if the current pNBO is the last one in this combination.
				eigenlists.append(eigenlist)
	for eigenlist in eigenlists: # Looping over degenerate groups.
		if len(eigenlist)==1: # Degenerate groups of only one pNBO, namely non-degenerate-pNBOs, do not need localizing.
			continue
		H=getMatrix(eigenlist)
		U=loc.oldPipekMezey(C@H,S,basis_indices_by_frag,"Lowdin",None) # C - NAO in AO basis; C@H - degenerate pNBO in AO basis
		H=H@U # Partially localizing the degenerate pNBOs.
		setMatrix(eigenlist,H)

	# Dividing pNBOs into pNHOs.
	for pnbo in pnbos: # Looping over pNBOs.
		for pnho in pnbo.pNHOs: # Looping over pNHOs of this pNBO.
			bif=basis_indices_by_frag[pnho.Fragment]
			pnho.Vector=np.zeros_like(pnbo.Vector)
			pnho.Vector[bif]=pnbo.Vector[bif] # Assigning the elements of this pNBO belonging to each fragment to the corresponding pNHO.

	# Identifying multi-electron bonds
	pnhos_frags=SortpNBO(pnhos,"Fragment") # Sorting pNHOs by fragments, [fragment : pNHO].
	for frag,pnhos_frag in pnhos_frags.items():
		for ipnho in range(len(pnhos_frag)):
			pnhoi=pnhos_frag[ipnho]
			if pnhoi.Multi==-1:
				continue
			vectori=pnhoi.Vector/np.linalg.norm(pnhoi.Vector)
			for jpnho in range(ipnho+1,len(pnhos_frag)):
				pnhoj=pnhos_frag[jpnho]
				if pnhoj.Multi==-1:
					continue
				vectorj=pnhoj.Vector/np.linalg.norm(pnhoj.Vector)
				if abs(np.inner(vectori,vectorj))>multi_thres:
					pnhoi.Multi=1
					pnhoi.Siblings.append(pnhoj)
					pnhoj.Multi=-1
					pnhoj.Siblings.append(pnhoi)
	for pnhoi in pnhos: # Averaging over sibling pNHOs by population.
		if pnhoi.Multi==1:
			for pnhoj in pnhoi.Siblings:
				pnhoi.Vector+=pnhoj.Vector*np.sign(np.inner(pnhoi.Vector,pnhoj.Vector))
	n_two_e_pnbos=len(pnbos)
	for ipnbo in range(n_two_e_pnbos):
		pnboi=pnbos[ipnbo]
		for jpnbo in range(ipnbo+1,n_two_e_pnbos):
			pnboj=pnbos[jpnbo]
			combined=False
			for pnhok in pnboi.pNHOs:
				for pnhol in pnboj.pNHOs:
					if pnhok in pnhol.Siblings:
						combined=True
						break
			if combined:
				pnmeb=None
				if pnboi.pNMEB!=None and pnboj.pNMEB!=None:
					continue
				elif pnboi.pNMEB!=None and pnboj.pNMEB==None:
					pnmeb=pnboi.pNMEB
					pnboj.pNMEB=pnmeb
					pnmeb.pNHOs+=[pnho for pnho in pnboj.pNHOs if pnho.Multi!=-1]
				elif pnboi.pNMEB==None and pnboj.pNMEB!=None:
					pnmeb=pnboj.pNMEB
					pnboi.pNMEB=pnmeb
					pnmeb.pNHOs+=[pnho for pnho in pnboi.pNHOs if pnho.Multi!=-1]
				else:
					pnmeb=pNBO()
					pnmeb.pNHOs=[pnho for pnho in pnboi.pNHOs+pnboj.pNHOs if pnho.Multi!=-1]
					pnboi.pNMEB=pnmeb
					pnboj.pNMEB=pnmeb
					pnbos.append(pnmeb)

	# Removing redundant pNHOs and pNBOs.
	delete_list=[ipnho for ipnho in range(len(pnhos)) if pnhos[ipnho].Multi==-1]
	'''
	for pnho in pnhos:
		if pnho.Multi==1:
			print(pnhos.index(pnho)+1,pnbos.index(pnho.pNBO)+1,end=' ')
			for pnhoj in pnho.Siblings:
				print(pnhos.index(pnhoj)+1,pnbos.index(pnhoj.pNBO)+1,end=' ')
			print('')
	'''
	delete_list.reverse()
	for ipnho in delete_list:
		pnhos.pop(ipnho)
	delete_list=[ipnbo for ipnbo in range(len(pnbos)) if pnbos[ipnbo].pNMEB!=None]
	delete_list.reverse()
	for ipnbo in delete_list:
		pnbos.pop(ipnbo)
	for pnbo in pnbos:
		if pnbo.Combination==None:
			comb_set=set()
			for pnho in pnbo.pNHOs:
				comb_set.add(pnho.Fragment)
			comb=tuple(comb_set)
			pnbo.Combination=comb

	# Orthogonalization of pNHOs and generation of NHOs.
	pnhos_frags=SortpNBO(pnhos,"Fragment") # Sorting pNHOs by fragments, [fragment : pNHO].
	for frag,pnhos_frag in pnhos_frags.items(): # Looping over fragments.
		Iblock=getMatrix(pnhos_frag) # The coefficients of the pNHOs of this fragment.
		Sblock=Iblock.T@Iblock # The overlap matrix of pNHOs in NAO basis set. NAOs are mutually orthonormal.
		Jblock=np.zeros([len(pnhos_frag),len(pnhos_frag)])
		for ipnho,pnho in zip(range(len(pnhos_frag)),pnhos_frag):
			Jblock[ipnho,ipnho]=np.linalg.norm(pnho.Vector) # Normalization factors of each pNHO.
		Oblock=Jblock@sl.sqrtm(np.linalg.inv(Jblock@Sblock@Jblock)) # Normalization-factor-weighted symmetric orthogonalization.
		Iblock=Iblock@Oblock
		setMatrix(pnhos_frag,Iblock)
		Jblock=np.diag(Iblock.T@P@Iblock)
		for ipnho in range(len(pnhos_frag)):
			pnhos_frag[ipnho].Occupancy=Jblock[ipnho] # Occupation numbers of NHOs.

	# Generation of NBO by diagonalization of the NHO-based density matrix.
	nhos=[]
	nbos=[]
	pnbos_combs=SortpNBO(pnbos,"Combination") # Sorting pNBOs by combinations, [combination : pNBO].
	for comb,pnbos_comb in pnbos_combs.items():
		for pnbo in pnbos_comb:
			nho_index_head=len(nhos)
			nhos_pnho=[]
			for pnho in pnbo.pNHOs:
				nho=NHO()
				nho.Fragment=pnho.Fragment
				nho.Occupancy=pnho.Occupancy
				nho.Vector=pnho.Vector
				nhos.append(nho)
				nhos_pnho.append(nho)
			nho_index_tail=len(nhos)
			nho_indices_pnbo=[i for i in range(nho_index_head,nho_index_tail)]
			Iblock=getMatrix(pnbo.pNHOs) # The NHOs of this pNBO.
			Pblock=Iblock.T@P@Iblock # The density matrix expressed by these NHOs.
			Nblock,Hblock=np.linalg.eigh(Pblock) # NBOs.
			Nblock=Nblock[::-1]
			Hblock=np.fliplr(Hblock)
			eigenlists=[]
			eigenlist=[]
			eigenvalue=114514
			for i in range(len(Nblock)):
				if not np.isclose(Nblock[i],eigenvalue,atol=deg_thres):
					eigenlists.append(eigenlist)
					eigenlist=[]
				eigenvalue=Nblock[i]
				eigenlist.append(i)
				if i+1==len(Nblock):
					eigenlists.append(eigenlist)
			for eigenlist in eigenlists:
				if len(eigenlist)<2:
					continue
				H=Hblock[:,eigenlist]
				U=loc.oldPipekMezey(C@Iblock@H,S,basis_indices_by_frag,"Lowdin",None)
				Hblock[:,eigenlist]=H@U
			for i in range(len(Nblock)):
				nbo=NBO()
				nbo.Combination=pnbo.Combination
				nbo.NHOs=nhos_pnho
				nbo.Occupancy=Nblock[i]
				nbo.Vector=np.zeros(nbasis)
				nbo.Vector[nho_indices_pnbo]=Hblock[:,i]
				nbo.Multi=any([pnho.Multi==1 for pnho in pnbo.pNHOs])
				nbos.append(nbo)

	return nhos,nbos


def NaturalBondOrbital(
		nao_mwfn,
		nao_info,
		frags = [],
		maxnfrags = -1,
		maxnnbos = -1,
		occ_thres = 0.95,
		multi_thres = 1,
		pdeg_thres = 1e-5,
		deg_thres = 0): # By default, every atom is a fragment, which is the case of NBO. By combining atoms into fragments one extends NBO to natural fragment bond orbital (NFBO).
	maxnfrags = nao_mwfn.getNumCenters() if maxnfrags == -1 else maxnfrags
	frags=[[i] for i in range(nao_mwfn.getNumCenters())] if frags == [] else frags
	basis_indices_by_center = nao_mwfn.Atom2BasisList()
	basis_indices_by_frag = []
	for frag in frags:
		basis_indices_this_fragment = []
		for icenter in frag:
			if icenter >= nao_mwfn.getNumCenters():
				raise RuntimeError("Atom index out of range!")
			basis_indices_this_fragment.extend(basis_indices_by_center[icenter])
		basis_indices_by_frag.append(basis_indices_this_fragment)
	nbasis = nao_mwfn.getNumBasis()
	if nao_mwfn.Overlap.shape != tuple([nbasis] * 2):
		nao_mwfn.Overlap = eint.PyscfOverlap(nao_mwfn, nao_mwfn)
	S = nao_mwfn.Overlap
	nho_mwfn = nao_mwfn.Clone()
	nbo_mwfn = nao_mwfn.Clone()
	print("Natural (fragment) bond orbitals:")
	for spin in nao_mwfn.getSpins():
		C = nao_mwfn.getCoefficientMatrix(spin)
		if maxnnbos == -1:
			maxnnbos = round(nao_mwfn.getNumElec(spin))
			if spin == 0:
				maxnnbos //= 2
		P = None
		match spin:
			case 0:
				P = nao_info["NAO_density_matrix"]
			case 1:
				P = nao_info["NAO_alpha_density_matrix"]
			case 2:
				P = nao_info["NAO_beta_density_matrix"]
		if spin == 0:
			occ_thres *= 2
			deg_thres *= 2
		nhos, nbos = generateNaturalBondOrbital(
				basis_indices_by_frag,
				P, C, S,
				maxnfrags, maxnnbos,
				occ_thres, multi_thres,
				pdeg_thres, deg_thres
		)
		output = "Spin " + str(spin) + "\n" # Printing NBO and NHO information
		nbos_combs = SortpNBO(nbos, "Combination") # Sorting NBOs by combinations, [combination : NBO].
		for comb, nbos_comb in nbos_combs.items():
			output += "Fragment combination " + str(comb) + "\n"
			for nbo in nbos_comb:
				inbo = nbos.index(nbo)
				occ = nbo.Occupancy.real
				nbo_mwfn.Orbitals[inbo + (nbasis if spin == 2 else 0)].Occ = occ
				output += "NBO_" + str(inbo + (nbasis if spin == 2 else 0)) + " (" + str(round(occ, 3)) + ")  ="
				for nho in nbo.NHOs:
					jnho = nhos.index(nho)
					frag = nho.Fragment
					occ = nho.Occupancy.real
					nho_mwfn.Orbitals[jnho + (nbasis if spin == 2 else 0)].Occ = occ
					coef = nbo.Vector[jnho]
					output += "  " + str(round(coef, 3)) \
							+ " * NHO_" + str(jnho + (nbasis if spin==2 else 0)) \
							+ " (" + str(round(occ, 3)) + ", F_" + str(frag)+ ")"
				output += "\n"
		print(output)
		I = np.zeros([nbasis, nbasis])
		I[:, :len(nhos)] = getMatrix(nhos)
		H = np.zeros([nbasis, nbasis])
		H[:, :len(nbos)] = getMatrix(nbos)
		nho_mwfn.setEnergy([0 for i in range(nbasis)], spin)
		nho_mwfn.setCoefficientMatrix(C @ I, spin)
		nbo_mwfn.setEnergy([0 for i in range(nbasis)], spin)
		nbo_mwfn.setCoefficientMatrix(C @ I @ H, spin)
	nbo_mwfn.Orthogonalize("GramSchmidt")
	return nho_mwfn, nbo_mwfn



import numpy as np
import scipy.interpolate as si


def MeshCubic(x1,x2,v1,v2,g1,g2,func,func_para,num):
    displacement=x2-x1
    length=np.linalg.norm(displacement)
    direction=displacement/length
    ls=np.linspace(0,length,num=num)
    xs=np.linspace(x1,x2,num=num)
    vs=np.array([v1]+[func(x,*func_para) for x in xs[1:-1]]+[v2])
    gl1=np.sum(g1*direction)
    gl2=np.sum(g2*direction)
    poly=si.CubicSpline(ls,vs,bc_type=((1,gl1),(1,gl2)))
    roots=poly.derivative().solve()
    possible_ls=[root for root in roots if root>0 and root<length]+[length]
    possible_xs=[x1+direction*l for l in possible_ls]
    possible_vs=[func(x,*func_para) for x in possible_xs[:-1]]+[v2]
    best_index=np.argmax(possible_vs)
    best_x=possible_xs[best_index]
    return best_x




import numpy as np
import scipy.linalg as sl
from . import DirectionSearch as dc
from . import LineSearch as ls


def Lehtola(C0,guess,func,jac,q,conv,func_para=(),jac_para=(),dir_optn={},line_optn={}):
    maxiter=1000
    np.random.seed(0)
    X=np.random.rand(C0.shape[1],C0.shape[1])
    X-=X.T
    X*=guess
    Ulast=np.zeros_like(X)+114514
    U=sl.expm(X)
    Llast=-1919810
    L=-1919810
    Glast=None
    G=None
    Hlast=None
    H=None
    for iiter in range(maxiter):
        if iiter>0:
            Ulast=U.copy()
            Llast=L
            Glast=G.copy()
            Hlast=H.copy()
            if iiter%C0.shape[1]==0:
                Glast=None
                Hlast=None
        L=func(U)
        G=jac(U) # Riemannian derivative
        H=dc.ConjugateGradient(Hlast,G,Glast,dir_optn.get("cgtype","PR"))
        wmax=np.max(np.abs(np.linalg.eigvalsh(1.j*H)))
        x1=np.zeros_like(H)
        x2=H*2*np.pi/wmax/q
        v1=L
        v2=func(sl.expm(x2)@U)
        g1=G
        g2=jac(sl.expm(x2)@U)
        func_for_line_search=lambda x:func(sl.expm(x)@U)
        best_x=ls.MeshCubic(x1,x2,v1,v2,g1,g2,func_for_line_search,(),line_optn.get("num",3))
        U=sl.expm(best_x)@U
        print("Iteration %d:  L = %f; Gmax = %f" % (iiter,L,np.max(np.abs(G))))
        if conv(U,L,G,Ulast,Llast,Glast):
            return U
    print("Convergence not fully achieved!")
    return U



import numpy as np
from Orbaplaw import Miscellany as mis
from Orbaplaw import Integrals as eint
from . import Lowdin
#from . import Mulliken

def PopulationAnalyzer(mo_mwfn, method = "Lowdin", space = "occ"):
	natoms = mo_mwfn.getNumCenters()
	S = mo_mwfn.Overlap
	if mo_mwfn.Overlap.shape != tuple([mo_mwfn.getNumBasis()] * 2):
		S = eint.PyscfOverlap(mo_mwfn, mo_mwfn)
	basis_indices_by_center = mo_mwfn.Atom2BasisList()
	for spin in mo_mwfn.getSpins():
		nocc = round(mo_mwfn.getNumElec(spin))
		if spin == 0:
			nocc //= 2
		orbital_range = []
		if type(space) is str:
			if space.upper() == "OCC":
				orbital_range = list(range(nocc))
		elif type(space) is list:
			orbital_range = space
		format_range = mis.FormatRange(orbital_range)
		C = mo_mwfn.getCoefficientMatrix(spin)[:, orbital_range]

		charge_type = ""
		Qs = []
		if "LOWDIN" in method.upper():
			charge_type = "Lowdin"
			Qs = Lowdin(C, S, basis_indices_by_center)
		elif "MULLIKEN" in method.upper():
			raise RuntimeError("Not implemented yet!")
			charge_type = "Mulliken"
			Qs = Mulliken(C, S, basis_indices_by_center)
		else:
			raise RuntimeError("Unrecognized charge type!")

		print("%s population on Spin %d Orbitals %s:" % ( charge_type, spin, format_range ))
		print("--------------------------------------------")
		print("| Index | Symbol | Population | Net Charge |")
		for iatom in range(natoms):
			symbol = mo_mwfn.Centers[iatom].getSymbol()
			population = np.sum(np.diag(Qs[iatom]))
			if spin == 0:
				population *= 2
			charge = mo_mwfn.Centers[iatom].Nuclear_charge - population
			print("| %5d | %6s | % 10f | % 10f |" % ( iatom, symbol, population, charge))
		print("--------------------------------------------")



import numpy as np
import Maniverse as mv

def FB_func(U, order, Wrefs, W2refSum):
	Ws = [ U.T @ Wref @ U for Wref in Wrefs]
	DiagWs = [ np.diag(np.diag(W)) for W in Ws ]
	W2Sum = U.T @ W2refSum @ U
	L = np.trace(W2Sum)
	for DiagW in DiagWs:
		L -= np.linalg.norm(DiagW) ** 2
	Ge = 2 * W2refSum @ U
	for Wref, DiagW in zip(Wrefs, DiagWs):
		Ge -= 4 * Wref @ U @ DiagW
	twoW2refSum = 2 * W2refSum
	fourWrefs = [ 4 * Wref for Wref in Wrefs ]
	eightWrefUs = [ 8 * Wref @ U for Wref in Wrefs ]
	UTWrefs = [ U.T @ Wref for Wref in Wrefs ]
	def He(v):
		return v
	if order == 2:
		def He(v):
			Hv = twoW2refSum @ v
			for fourWref, DiagW, eightWrefU, UTWref in zip(fourWrefs, DiagWs, eightWrefUs, UTWrefs):
				Hv -= fourWref @ v @ DiagW + eightWrefU @ np.diag(np.diag(UTWref @ v))
			return Hv
	return L, Ge, He

def FosterBoys(Wrefs, W2refSum):
	M = mv.Orthogonal(np.eye(Wrefs[0].shape[0]), True)
	def func(U, order):
		L, Ge, He = FB_func(U, order, Wrefs, W2refSum)
		return L, Ge, He
	L = 0
	tr_setting = mv.TrustRegionSetting()
	tol0 = 1e-8 * M.getDimension()
	tol1 = 1e-6 * M.getDimension()
	tol2 = 10
	mv.TrustRegion(
			func, tr_setting, (tol0, tol1, tol2),
			0.001, 1, 1000, L, M, 1
	)
	return M.P



from Orbaplaw.NaturalBondOrbitalMethods import NaturalAtomicOrbital
from Orbaplaw.NaturalBondOrbitalMethods import NaturalBondOrbital
from Orbaplaw.NaturalBondOrbitalMethods import PrincipalInteractingOrbital

generateNaturalAtomicOrbital=NaturalAtomicOrbital.generateNaturalAtomicOrbital
NaturalAtomicOrbital=NaturalAtomicOrbital.NaturalAtomicOrbital
generateNaturalBondOrbital=NaturalBondOrbital.generateNaturalBondOrbital
NaturalBondOrbital=NaturalBondOrbital.NaturalBondOrbital
generatePrincipalInteractingOrbital=PrincipalInteractingOrbital.generatePrincipalInteractingOrbital
PrincipalInteractingOrbital=PrincipalInteractingOrbital.PrincipalInteractingOrbital



import re
import argparse
import libmwfn as lm
from Orbaplaw import Population as pop
from Orbaplaw import Localization as loc
from Orbaplaw import OrbitalAlignment as oa
from Orbaplaw import NaturalBondOrbitalMethods as nbo

def parse_range(value):
	"""Parse a string like '3-10,12,17-20' into a list of integers."""
	ranges = []
	parts = value.split(',')
	
	for part in parts:
		if '-' in part:
			# Handle ranges like "3-10"
			subparts = part.split('-')
			if len(subparts) != 2:
				raise argparse.ArgumentTypeError(f"Invalid range: '{part}'")
			try:
				start = int(subparts[0])
				end = int(subparts[1])
			except ValueError:
				raise argparse.ArgumentTypeError(f"Non-integer in range: '{part}'")
			if start > end:
				raise argparse.ArgumentTypeError(f"Invalid range order: '{part}'")
			ranges.extend(range(start, end + 1))  # Inclusive of end
		else:
			# Handle single numbers like "12"
			try:
				num = int(part)
				ranges.append(num)
			except ValueError:
				raise argparse.ArgumentTypeError(f"Invalid number: '{part}'")
	return ranges

def parse_alpha_range(value):
	if bool(re.search(r'[a-zA-Z]', value)):
		return value
	else:
		return parse_range(value)

def main():
	parser = argparse.ArgumentParser(description = "Command-line tool for Orbaplaw")
	subparsers = parser.add_subparsers(
			title = "Job types",
			dest = "job_type",
			required = True
	)

	# Parser for population analysis
	parser_pop = subparsers.add_parser(
			"pop",
			help = "Population Analysis"
	)
	parser_pop.add_argument(
			"-i", "--input",
			required = True,
			help = "Mwfn file (Required)"
	)
	parser_pop.add_argument(
			"--charge",
			default = "Lowdin",
			help = "Charge type (Default: %(default)s)"
	)
	parser_pop.add_argument(
			"--space",
			default = "occ",
			type = parse_alpha_range,
			help = "Orbital space to analyze ('occ', 'vir', '0-2,4,6,8-10') (Default: %(default)s)"
	)

	# Parser for localization
	parser_loc = subparsers.add_parser(
			"loc",
			help = "Orbital localization"
	)
	parser_loc.add_argument(
			"-i", "--input",
			required = True,
			help = "Original mwfn file (Required)"
	)
	parser_loc.add_argument(
			"-o", "--output",
			required = True,
			help = "Mwfn file for exported localized orbitals (Required)"
	)
	parser_loc.add_argument(
			"--method",
			default = "PipekMezey-Lowdin",
			help = "Localization method ('PipekMezey[-Lowdin/-Mulliken]', 'Foster-Boys', 'Orbitalet[#0.7959/#gamma_e]') (Default: %(default)s)"
	)
	parser_loc.add_argument(
			"--space",
			default = "occ",
			type = parse_alpha_range,
			help = "Orbital space to localize ('occ', 'vir', '0-2,4,6,8-10') (Default: %(default)s)"
	)

	# Parser for fragment alignment
	parser_famo = subparsers.add_parser(
			"famo",
			help = "Fragment aligned molecular orbital"
	)
	parser_famo.add_argument(
			"-i", "--input",
			required = True,
			help = "Original mwfn file of the whole molecule (Required)"
	)
	parser_famo.add_argument(
			"-f", "--fragments",
			required = True,
			nargs = '+',
			help = "Fragment mwfn file(s)"
	)
	parser_famo.add_argument(
			"-o", "--output",
			required = True,
			help = "Mwfn file for exported FAMOs (Required)"
	)
	parser_famo.add_argument(
			"--diagmat",
			default = False,
			type = bool,
			help = "Whether to diagonalize the matched space (Default: %(default)s)"
	)
	parser_famo.add_argument(
			"--diagmis",
			default = True,
			type = bool,
			help = "Whether to diagonalize the mismatched space (Default: %(default)s)"
	)

	# Parser for spin alignment
	parser_sno = subparsers.add_parser(
			"sno",
			help = "Spin natural orbital"
	)
	parser_sno.add_argument(
			"-i", "--input",
			required = True,
			help = "Original mwfn file (Required)"
	)
	parser_sno.add_argument(
			"-o", "--output",
			required = True,
			help = "Mwfn file for exported SNOs (Required)"
	)
	parser_sno.add_argument(
			"--diagmat",
			default = False,
			type = bool,
			help = "Whether to diagonalize the matched space (Default: %(default)s)"
	)
	parser_sno.add_argument(
			"--diagmis",
			default = True,
			type = bool,
			help = "Whether to diagonalize the mismatched space (Default: %(default)s)"
	)

	# Parser for natural atomic orbitals
	parser_nao = subparsers.add_parser(
			"nao",
			help = "Natural atomic orbital"
	)
	parser_nao.add_argument(
			"-i", "--input",
			required = True,
			help = "Original mwfn file (Required)"
	)
	parser_nao.add_argument(
			"-o", "--output",
			required = True,
			help = "Mwfn file for exported NAOs (Required)"
	)

	# Parser for principal interacting orbitals
	parser_pio = subparsers.add_parser(
			"pio",
			help = "Principal interacting orbital"
	)
	parser_pio.add_argument(
			"-i", "--input",
			required = True,
			help = "Original mwfn file (Required)"
	)
	parser_pio.add_argument(
			"--fragments",
			required = True,
			nargs = 2,
			help = "Fragmentation scheme (e.g., '0-2,4 6,8-10') (Required)"
	)
	parser_pio.add_argument(
			"-o", "--output",
			required = True,
			nargs = 2,
			help = "Mwfn file for exported PIOs and PIMOs, respectively (Required)"
	)

	# Parser for natural bond orbitals
	parser_nbo = subparsers.add_parser(
			"nbo",
			help = "Natural (fragment) bond orbital"
	)
	parser_nbo.add_argument(
			"-i", "--input",
			required = True,
			help = "Original mwfn file (Required)"
	)
	parser_nbo.add_argument(
			"-o", "--output",
			required = True,
			nargs = 2,
			help = "Mwfn file for exported NHOs and NBOs, respectively (Required)"
	)
	parser_nbo.add_argument(
			"--fragments",
			nargs = '*',
			default = [],
			help = "Fragmentation scheme (e.g., '0-2,4 6,8-10 ...') (Default: _EMPTY_ -> each atom is a fragment)"
	)
	parser_nbo.add_argument(
			"--maxnfrags",
			default = -1,
			type = int,
			help = "Maximum number of fragments involved in a bonding scheme (Default: %(default)s -> no limit)"
	)
	parser_nbo.add_argument(
			"--maxnnbos",
			default = -1,
			type = int,
			help = "Maximum number of bonding schemes to find (Default: %(default)s -> no limit)"
	)
	parser_nbo.add_argument(
			"--occ_thres",
			default = 0.95,
			type = float,
			help = "Occupation threshold for recognizing pNFBOs (ranging from 0 to 1) (Default: %(default)s)"
	)
	parser_nbo.add_argument(
			"--multi_thres",
			default = 1.,
			type = float,
			help = "Occupation threshold for recognizing multi-electron bonds (ranging from 0 to 1) (Default: %(default)s)"
	)
	parser_nbo.add_argument(
			"--pdeg_thres",
			default = 1e-5,
			type = float,
			help = "Occupation difference threshold for recognizing degenerate pNFBOs (Default: %(default)s)"
	)
	parser_nbo.add_argument(
			"--deg_thres",
			default = 0.,
			type = float,
			help = "Occupation difference threshold for recognizing degenerate NFBOs (Default: %(default)s)"
	)


	args = parser.parse_args()


	# Running population analysis
	if args.job_type == "pop":
		mwfn = lm.Mwfn(args.input)
		pop.PopulationAnalyzer(mwfn, args.charge, args.space)

	# Running localization
	if args.job_type == "loc":
		orig_mwfn = lm.Mwfn(args.input)
		loc_mwfn = loc.Localizer(orig_mwfn, args.method, args.space)
		loc_mwfn.Export(args.output)

	# Running fragment alignment
	if args.job_type == "famo":
		whole_mwfn = lm.Mwfn(args.input)
		frag_mwfns = []
		for frag_mwfn in args.fragments:
			frag_mwfns.append(mlm.Mwfn(frag_mwfn))
		famo_mwfn = oa.FragmentAlignment(whole_mwfn, frag_mwfns, args.diagmat, args.diagmis)
		famo_mwfn.Export(args.output)

	# Running spin alignment
	if args.job_type == "sno":
		orig_mwfn = lm.Mwfn(args.input)
		sno_mwfn = oa.SpinAlignment(orig_mwfn, args.diagmat, args.diagmis)
		sno_mwfn.Export(args.output)

	# Running natural atomic orbitals
	if args.job_type == "nao":
		orig_mwfn = lm.Mwfn(args.input)
		nao_mwfn, _ = nbo.NaturalAtomicOrbital(orig_mwfn)
		nao_mwfn.Export(args.output)

	# Running principal interacting orbitals
	if args.job_type == "pio":
		orig_mwfn = lm.Mwfn(args.input)
		nao_mwfn, nao_info = nbo.NaturalAtomicOrbital(orig_mwfn)
		frags = []
		for frag in args.fragments:
			frags.append(parse_range(frag))
		pio_mwfn, pimo_mwfn = nbo.PrincipalInteractingOrbital(nao_mwfn, nao_info, frags)
		pio_mwfn.Export(args.output[0])
		pimo_mwfn.Export(args.output[1])

	# Running natural bond orbitals
	if args.job_type == "nbo":
		orig_mwfn = lm.Mwfn(args.input)
		nao_mwfn, nao_info = nbo.NaturalAtomicOrbital(orig_mwfn)
		frags = []
		for frag in args.fragments:
			frags.append(parse_range(frag))
		nho_mwfn, nbo_mwfn = nbo.NaturalBondOrbital(nao_mwfn, nao_info, frags, args.maxnfrags, args.maxnnbos, args.occ_thres, args.multi_thres, args.pdeg_thres, args.deg_thres)
		nho_mwfn.Export(args.output[0])
		nbo_mwfn.Export(args.output[1])



import numpy as np
from pyscf import gto


def PyscfDecompose(mwfn):
	atom_string = ""
	for iatom, atom in enumerate(mwfn.Centers):
		atom_string += atom.getSymbol() + str(iatom) + " " + str(atom.Coordinates[0]) + " " + str(atom.Coordinates[1]) + " " + str(atom.Coordinates[2]) + "\n"
	mol_list = []
	cart_list = []
	head_list = []
	tail_list = []
	center_list = mwfn.Shell2Atom()
	nbasis = 0
	charge = round(mwfn.getCharge())
	spin = round(mwfn.getNumElec(1) - mwfn.getNumElec(2))
	shell2atom = mwfn.Shell2Atom()
	for ishell in range(mwfn.getNumShells()):
		shell = mwfn.getShell(ishell)
		mol = gto.Mole()
		mol_list.append(mol)
		cart_list.append(shell.Type >= 2)
		head_list.append(nbasis)
		nbasis += shell.getSize()
		tail_list.append(nbasis)
		jcenter = shell2atom[ishell]
		mol.atom = atom_string
		mol.basis = {
				mwfn.Centers[jcenter].getSymbol() + str(jcenter):[ [abs(shell.Type)] + [(shell.Exponents[j], shell.Coefficients[j]) for j in range(shell.getNumPrims())] ]
		}
		mol.unit = 'B'
		mol.charge = charge
		mol.spin = spin
		mol.build()
	return mol_list, cart_list, head_list, tail_list

def PyscfOverlap(mwfn1, mwfn2):
	overlap = np.zeros([mwfn1.getNumBasis(), mwfn2.getNumBasis()])
	mol_list1, cart_list1, head_list1, tail_list1 = PyscfDecompose(mwfn1)
	mol_list2, cart_list2, head_list2, tail_list2 = PyscfDecompose(mwfn2)
	for ishell in range(len(mol_list1)):
		imol = mol_list1[ishell]
		icart = cart_list1[ishell]
		ihead = head_list1[ishell]
		itail = tail_list1[ishell]
		for jshell in range(len(mol_list2)):
			jmol = mol_list2[jshell]
			jcart = cart_list2[jshell]
			jhead = head_list2[jshell]
			jtail = tail_list2[jshell]
			if icart or jcart:
				raise RuntimeError("Currently the integrals among cartesian basis functions (l >= 2) are not supported!")
			overlap[ihead:itail, jhead:jtail] = gto.intor_cross("int1e_ovlp_sph", imol, jmol)
	return overlap

# Below is an equivalent but more flexible realization of the function above.

nagging='''
mol_list1, cart_list1, head_list1, tail_list1 = PyscfDecompose(mwfn1)
mol_list2, cart_list2, head_list2, tail_list2 = PyscfDecompose(mwfn2)
for ishell in range(len(mol_list1)):
	imol = mol_list1[ishell]
	icart = cart_list1[ishell]
	ihead = head_list1[ishell]
	itail = tail_list1[ishell]
	for jshell in range(len(mol_list2)):
		jmol = mol_list2[jshell]
		jcart = cart_list2[jshell]
		jhead = head_list2[jshell]
		jtail = tail_list2[jshell]
		if icart or jcart:
			raise RuntimeError("Currently the integrals among cartesian basis functions (l >= 2) are not supported!")
__replacement__
'''

def PyscfOverlap(mwfn1, mwfn2):
	overlap = np.zeros([mwfn1.getNumBasis(), mwfn2.getNumBasis()])
	replacement = '''
		overlap[ihead:itail, jhead:jtail] = gto.intor_cross("int1e_ovlp_sph", imol, jmol)
	'''
	exec(nagging.replace("__replacement__", replacement))
	return overlap

def PyscfDipole(mwfn1, mwfn2):
	X = np.zeros([mwfn1.getNumBasis(), mwfn2.getNumBasis()])
	Y = np.zeros([mwfn1.getNumBasis(), mwfn2.getNumBasis()])
	Z = np.zeros([mwfn1.getNumBasis(), mwfn2.getNumBasis()])
	replacement = '''
		X[ihead:itail, jhead:jtail], Y[ihead:itail, jhead:jtail], Z[ihead:itail, jhead:jtail] = gto.intor_cross("int1e_r_sph", imol, jmol, comp=3)
	'''
	exec(nagging.replace("__replacement__", replacement))
	return X, Y, Z

def PyscfQuadrupole(mwfn1, mwfn2):
	XX = np.zeros([mwfn1.getNumBasis(), mwfn2.getNumBasis()])
	XY = np.zeros([mwfn1.getNumBasis(), mwfn2.getNumBasis()])
	XZ = np.zeros([mwfn1.getNumBasis(), mwfn2.getNumBasis()])
	YY = np.zeros([mwfn1.getNumBasis(), mwfn2.getNumBasis()])
	YZ = np.zeros([mwfn1.getNumBasis(), mwfn2.getNumBasis()])
	ZZ = np.zeros([mwfn1.getNumBasis(), mwfn2.getNumBasis()])
	replacement = '''
		XX[ihead:itail, jhead:jtail], XY[ihead:itail, jhead:jtail], XZ[ihead:itail, jhead:jtail], _, YY[ihead:itail, jhead:jtail], YZ[ihead:itail, jhead:jtail], _, _, ZZ[ihead:itail, jhead:jtail] = gto.intor_cross("int1e_rr_sph", imol, jmol, comp=9)
	'''
	exec(nagging.replace("__replacement__", replacement))
	return XX, XY, XZ, YY, YZ, ZZ



import numpy as np
import xml.etree.ElementTree as et
import os
import shutil as st
import copy as cp

class QeKPoint:
    x=None
    e=None
    npws=None
    mill=None
    ngw=None
    igwx=None
    evc=None

class QeWaveFunction:
    # Wavefunction info produced by QE.
    # For more information, see
    # https://mattermodeling.stackexchange.com/questions/9149/how-to-read-qes-wfc-dat-files-with-python
    # and 
    # https://gitlab.com/QEF/q-e/-/wikis/Developers/Format-of-data-files
    # .

    xml=None
    chargedat=None
    ispin=None
    gamma_only=None
    scalef=None
    npol=None
    nbnd=None
    b=None
    evc=None
    kpoints=None

    def __init__(self,folder):
        # Initializing a QeWaveFunction object from a *.save folder.

        # data-file-schema.xml
        self.xml=et.parse(folder+"/data-file-schema.xml")
        self.kpoints=[]
        for i in self.xml.getroot().find("output").find("band_structure").findall("ks_energies"):
            self.kpoints.append(QeKPoint())
            self.kpoints[-1].npws=np.array(i.find("npw").text.split(),'int32')[0]
            self.kpoints[-1].e=np.array(i.find("eigenvalues").text.split(),'float64')

        # charge-density.dat
        with open(folder+"/charge-density.dat","rb") as f:
            self.chargedat=f.read()

        # wfc*.dat
        for ik,k in zip(range(len(self.kpoints)),self.kpoints):
            with open(folder+"/wfc"+str(ik+1)+".dat",'rb') as f:

                f.seek(4)
                np.fromfile(f,dtype='int32',count=1)[0]
                k.x=np.fromfile(f,dtype='float64',count=3)
                self.ispin=np.fromfile(f,dtype='int32',count=1)[0]
                self.gamma_only=bool(np.fromfile(f,dtype='int32',count=1)[0])
                self.scalef=np.fromfile(f,dtype='float64',count=1)[0]
                f.seek(4,1)

                f.seek(4,1)
                k.ngw=np.fromfile(f,dtype='int32',count=1)[0]
                k.igwx=np.fromfile(f,dtype='int32',count=1)[0]
                self.npol=np.fromfile(f,dtype='int32',count=1)[0]
                self.nbnd=np.fromfile(f,dtype='int32',count=1)[0]
                f.seek(4,1)

                f.seek(4,1)
                self.b=np.zeros([3,3])
                self.b[0,:]=np.fromfile(f,dtype='float64',count=3)
                self.b[1,:]=np.fromfile(f,dtype='float64',count=3)
                self.b[2,:]=np.fromfile(f,dtype='float64',count=3)
                f.seek(4,1)

                f.seek(4,1)
                k.mill=np.fromfile(f,dtype='int32',count=3*k.igwx)
                k.mill=k.mill.reshape([k.igwx,3]).T
                f.seek(4,1)

                k.evc=np.zeros((self.npol*k.igwx,self.nbnd),dtype="complex128")
                for i in range(self.nbnd):
                    f.seek(4,1)
                    k.evc[:,i]=np.fromfile(f,dtype='complex128',count=self.npol*k.igwx)
                    f.seek(4,1)

    def Export(self,folder):
        if (os.path.exists(folder)):
            st.rmtree(folder)
        os.mkdir(folder)

        # data-file-schema.xml
        newxml=cp.deepcopy(self.xml)
        for i,k in zip(
                newxml.getroot().find("output").find("band_structure").findall("ks_energies"),
                self.kpoints):
            i.find("eigenvalues").text=np.array2string(k.e)[1:-1]
        newxml.write(folder+"/data-file-schema.xml")

        # charge-density.dat
        with open(folder+"/charge-density.dat","wb") as f:
            f.write(self.chargedat)
        '''
        with open(folder+"/charge-density.dat","wb") as f:
            f.write(np.array([12],dtype='int32').tobytes('F'))
            f.write(np.array([self.gamma_only],dtype='int32').tobytes('F'))
            f.write(np.array([self.igwx],dtype='int32').tobytes('F'))
            f.write(np.array([self.ispin],dtype='int32').tobytes('F'))
            f.write(np.array([12],dtype='int32').tobytes('F'))

            f.write(np.array([self.b.size*8],dtype='int32').tobytes('F'))
            f.write(np.array(self.b[0,:],dtype='float64').tobytes('F'))
            f.write(np.array(self.b[1,:],dtype='float64').tobytes('F'))
            f.write(np.array(self.b[2,:],dtype='float64').tobytes('F'))
            f.write(np.array([self.b.size*8],dtype='int32').tobytes('F'))

            f.write(np.array([self.mill.size*4],dtype='int32').tobytes('F'))
            f.write(np.array(self.mill,dtype='int32').tobytes('F'))
            f.write(np.array([self.mill.size*4],dtype='int32').tobytes('F'))

            f.write(np.array([self.igwx*4],dtype='int32').tobytes('F'))
            for i in range(self.igwx):
                f.write(np.array([self.igwx*4],dtype='int32').tobytes('F'))
            f.write(np.array([self.igwx*4],dtype='int32').tobytes('F'))
        '''

        # wfc*.dat
        for ik,k in zip(range(len(self.kpoints)),self.kpoints):
            with open(folder+"/wfc"+str(ik+1)+".dat",'wb') as f:
                f.write(np.array([44],dtype='int32').tobytes('F')) # Fortran feature: The byte size of the data block is shown at its head and tail.
                f.write(np.array([ik+1],dtype='int32').tobytes('F')) # Using fortran byte format.
                f.write(np.array(k.x,dtype='float64').tobytes('F')) # Reiterating the dtype is necessary because, say, there are int8, int32 and int64 but QE recognizes int32 only.
                f.write(np.array([self.ispin],dtype='int32').tobytes('F'))
                f.write(np.array([self.gamma_only],dtype='int32').tobytes('F'))
                f.write(np.array([self.scalef],dtype='float64').tobytes('F'))
                f.write(np.array([44],dtype='int32').tobytes('F'))

                f.write(np.array([16],dtype='int32').tobytes('F'))
                f.write(np.array([k.ngw],dtype='int32').tobytes('F'))
                f.write(np.array([k.igwx],dtype='int32').tobytes('F'))
                f.write(np.array([self.npol],dtype='int32').tobytes('F'))
                f.write(np.array([self.nbnd],dtype='int32').tobytes('F'))
                f.write(np.array([16],dtype='int32').tobytes('F'))

                f.write(np.array([self.b.size*8],dtype='int32').tobytes('F'))
                f.write(np.array(self.b[0,:],dtype='float64').tobytes('F'))
                f.write(np.array(self.b[1,:],dtype='float64').tobytes('F'))
                f.write(np.array(self.b[2,:],dtype='float64').tobytes('F'))
                f.write(np.array([self.b.size*8],dtype='int32').tobytes('F'))

                f.write(np.array([k.mill.size*4],dtype='int32').tobytes('F'))
                f.write(np.array(k.mill,dtype='int32').tobytes('F'))
                f.write(np.array([k.mill.size*4],dtype='int32').tobytes('F'))

                for i in range(self.nbnd):
                    f.write(np.array([k.evc[:,i].size*16],dtype='int32').tobytes('F'))
                    f.write(np.array(k.evc[:,i],dtype='complex128').tobytes('F'))
                    f.write(np.array([k.evc[:,i].size*16],dtype='int32').tobytes('F'))



def CompareQeWF(wf1,wf2): # Comparing the two QeWaveFunction objects.
    print('ispin',wf1.ispin-wf2.ispin)
    print('gamma_only',wf1.gamma_only is wf2.gamma_only)
    print('scalef',wf1.scalef-wf2.scalef)
    print('npol',wf1.npol-wf2.npol)
    print('nbnd',wf1.nbnd-wf2.nbnd)
    print('b',wf1.b-wf2.b)
    x=e=npws=mill=ngw=igwx=evc=0
    for i,j in zip(wf1.kpoints,wf2.kpoints):
        x+=np.linalg.norm(i.x-j.x)
        e+=np.linalg.norm(i.e-j.e)
        npws+=np.linalg.norm(i.npws-j.npws)
        mill+=np.linalg.norm(i.mill-j.mill)
        ngw+=np.linalg.norm(i.ngw-j.ngw)
        igwx+=np.linalg.norm(i.igwx-j.igwx)
        evc+=np.linalg.norm(i.evc-j.evc)
    print("x",x)
    print("e",e)
    print("npws",npws)
    print("mill",mill)
    print("ngw",ngw)
    print("igwx",igwx)
    print("evc",evc)


if __name__=='__main__':
    # Checking whether parsing and exporting is correct.
    from sys import argv
    wf1=QeWaveFunction(argv[1]) # Creating a QeWaveFunction object from a *.save folder.
    wf1.Export(argv[2]) # Exporting it to another *.save folder.
    wf2=QeWaveFunction(argv[2]) # Reading the two wfc*.dat file and checking whether they are the same.
    CompareQeWF(wf1,wf2)



import numpy as np


def ConjugateGradient(lasth,g,lastg,cgtype):
    gamma=0
    if lasth is None or lastg is None:
        return g
    if cgtype=="FR":
        gamma=np.linalg.norm(g)/np.linalg.norm(lastg)
    elif cgtype=="PR":
        gamma=np.sum(g*(g-lastg))/np.linalg.norm(lastg)
    h=g+gamma*lasth
    if np.sum(h*g)<0:
        h=g
    return h



import numpy as np
import Maniverse as mv
from . import FB_func
from . import Fock_func

def Orbitalet(Wrefs, W2refSum, Eref, gamma_e):
	M = mv.Orthogonal(np.eye(len(Eref)), True)
	Scale1 = 1. - gamma_e
	Scale2 = gamma_e * 1000
	def func(U, order):
		L1, Ge1, He1 = FB_func(U, order, Wrefs, W2refSum)
		L2, Ge2, He2 = Fock_func(U, order, Eref)
		L = Scale1 * L1 + Scale2 * L2
		Ge = Scale1 * Ge1 + Scale2 * Ge2
		def He(v):
			return Scale1 * He1(v) + Scale2 * He2(v)
		return L, Ge, He
	L = 0
	tr_setting = mv.TrustRegionSetting()
	tol0 = 1e-8 * M.getDimension()
	tol1 = 1e-6 * M.getDimension()
	tol2 = 10
	mv.TrustRegion(
			func, tr_setting, (tol0, tol1, tol2),
			0.001, 1, 1000, L, M, 1
	)
	return M.P



from Orbaplaw.Localization import PipekMezey
PM_func = PipekMezey.PM_func
PM_jac = PipekMezey.PM_jac
oldPipekMezey = PipekMezey.oldPipekMezey
PipekMezey = PipekMezey.PipekMezey

from Orbaplaw.Localization import FosterBoys
FB_func = FosterBoys.FB_func
FosterBoys = FosterBoys.FosterBoys

from Orbaplaw.Localization import Fock
Fock_func = Fock.Fock_func
Fock = Fock.Fock

from Orbaplaw.Localization import Orbitalet
Orbitalet = Orbitalet.Orbitalet

from Orbaplaw.Localization import Localizer
Localizer = Localizer.Localizer



import numpy as np

def OrbitalAlignment(A, B, S, epsilon, diagmat, diagmis):
	# A - the coefficient matrix of the whole molecule
	# B - the coefficient matrices of the fragments
	# S - the overlap matrix of the two basis sets, row -> basis of A, column -> basis of B
	# epsilon - the orbital energies of the whole molecule
	# diagmat - whether to diagonalize the matched space
	# diagmis - whether to diagonalize the mismatched space
	m = A.shape[0] # The number of A's basis functions
	n = A.shape[1] # The number of A's occupied orbitals
	p = B.shape[0] # The number of B's basis functions
	q = B.shape[1] # The number of B's occupied orbitals
	# m > n, p > q.
	# In the case of the same cutoff and the same lattice parameters, m = p and n > q.
	I = np.zeros([q, n])
	for i in range(min(q, n)):
		I[i, i] = 1.
		U, Sigma_or_E, VH = np.linalg.svd((A.conj().T @ B @ I) if S is None else (A.conj().T @ S @ B @ I))
	X = U @ VH
	C = A @ X
	Sigma_or_E = np.zeros(n)
	SS = C.conj().T @ B if S is None else C.conj().T @ S @ B
	for i in range(q):
		Sigma_or_E[i] = np.max(SS[i, :])

	Y = np.eye(n)
	if epsilon is not None:
		Fo = X.conj().T @ np.diag(epsilon) @ X
		if diagmat:
			Fmat = Fo[0 : q, 0 : q]
			Sigma_or_E[0 : q], Y[0 : q, 0 : q] = np.linalg.eigh(Fmat)
		if diagmis:
			Fmis = Fo[q : n, q : n]
			Sigma_or_E[q : n], Y[q : n, q : n] = np.linalg.eigh(Fmis)
	return A @ X @ Y, Sigma_or_E



from Orbaplaw.Optimization import DirectionSearch
ConjugateGradient=DirectionSearch.ConjugateGradient

from Orbaplaw.Optimization import LineSearch
MeshCubic=LineSearch.MeshCubic

from Orbaplaw.Optimization import UnitaryOptimizer
Lehtola=UnitaryOptimizer.Lehtola
